#!/usr/bin/env python
"""\
Simple HTTP server that requests for a automatically generated token before
access to the data is granted.

The use case for this is:
You want to quickly share a file with someone. You have set up a network
connecion (LAN, cross over cable, Wireless Ad-Hoc network, whatever).
But you don't want to set up an account so that for example ssh/scp login
is possible. You don't want to do a complex export of a directory with
SMB or NFS.

With this script you can easily share all files in the folder its started.
The peer onlys need a standard webbrowser, your IP and the token.

A token is automatically generated for each peer that wants to access the
data. It displays a challange to the peer, which he has to tell to the
person running the server. He can then decide if he wants to grant access
and tell the peer the token.
As for every connection a new challange/response is calculated it does not
matter if a 3rd person listens to the two exchanging the keys. When the 3rd
person tries to fetch files it will get a new session with different
challange/response.

However it does not have strong security so shut down the server as soon as
the files are exchanged.

(C) 2007 - Chris Liechti <cliechti@gmx.net>
"""

import SimpleHTTPServer
import random
import time
import sys
import os
import Cookie
import urlparse
import cgi
import urllib
import stat
import posixpath
import BaseHTTPServer
import SocketServer
from cStringIO import StringIO

# -----------------------------------------------------------------------------

righthand = '23456qwertasdfgzxcvb'
lefthand = '789yuiophjknm'
characters = (righthand, lefthand)

def token(length=4):
    """generate a simple token of given length"""
    rnd = random.Random()
    while length:
        yield rnd.choice(characters[length & 1])
        length -= 1

# -----------------------------------------------------------------------------

unitprefixes = ['', 'Ki', 'Mi', 'Gi', 'Ti']
def bytes2string(num):
    """give a nice formated byte count. see http://en.wikipedia.org/wiki/Kibi"""
    prefix = 0
    while num >= 1024*1024:
        prefix += 1
        num /= 1024
    num = float(num)
    if num >= 1024:
        prefix += 1
        num /= 1024
    if prefix > 0:
        try:
            str_p = unitprefixes[prefix]
        except IndexError:
            str_p = "* 2**%d" % (prefix*10)
        return "%.2f %sB" % (num, str_p)
    else:
        return "%d B" % (num)

# -----------------------------------------------------------------------------
def format_time(now):
    year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
    return "%04d-%02d-%02d %02d:%02d:%02d" % (year, month, day, hh, mm, ss)

# -----------------------------------------------------------------------------

class Session(object):
    """\
    Keep track of a session. A session has 3 random strings:
    - A long identification. This is set as cookie, its never presented to the
      user or asked for.
    - A short challange which the peer must read and tell the server admin.
    - A short response which the server admin tells the peer which in turn
      enters it in the login form.
    """
    def __init__(self):
        """create a new session with an id, challange and response. its
           creation and expriation time is stored"""
        self.created = time.time()
        self.client_address = None
        self.max_age = 900 #15 min
        self.id = '%s-%s' % (self.created, ''.join(token(24)))
        self.challange = ''.join(token(4))
        self.response = ''.join(token(4))
        self.expires = self.created + self.max_age
        self.access_granted = False
        self.trail = []
        print "New Session: %s -> %s" % (self.challange, self.response)

    def is_expired(self):
        """returns true if the session is too old and thus invalid"""
        return time.time() > self.expires

    def prolong(self):
        """prolong the validity of this session. usually used when the
           session is kept active through activity"""
        # only authenticated sessions are prolonged
        if self.access_granted:
            self.expires = time.time() + self.max_age

    def track(self, location, message):
        """make an entry to the trail"""
        self.trail.append((time.time(), location, message))

    def __str__(self):
        """a textual representation"""
        return "Session[id:%r challange:%r response:%r]" % (self.id, self.challange, self.response)

# -----------------------------------------------------------------------------
css_tt = 'font-family:monospace;'
css_challange = 'font-family:monospace;border-color:red;border-style:solid;border-width:1px;'
css_response = 'font-family:monospace;border-color:green;border-style:solid;border-width:1px;'
css_border = 'border-width:2px;border-style:solid;border-color:black'

# The raw HTML of the login form. A few variables are used:
# 'challange', 'response', 'message'
page_login_form = """\
<html>
    <head>
        <title>Simple Share</title>
    </head>
    <body
      style="font-family:sans-serif;color:black;background:#bbddff"
      onload="javascript:document.login.response.focus()"
    >
        <center>
            <div style="color:white;background:#556688;font-size:150%%">
                Simple Share
            </div>
            <h1>%(message)s</h1>
            <form name="login" method="get" action="/login">
                <div>The challange is
                    <strong style="border-color:red;border-style:solid;border-width:2px;font-size:150%%">%(challange)s</strong>,
                    enter the response here
                    <input name="response" size="4" maxlength="4" style="border-color:green;border-style:solid;border-width:2px;font-size:150%%" />
                    <input type="hidden" name="redirect" value="%(redirect)s" />
                </div>
                <div>Note: Cookies must be enabled.</div>
                <div>&nbsp;</div>
                <div style="color:white;background:#556688;font-size:150%%">
                    <input type="submit" value="Login" style="border-color:black;border-style:solid;border-width:1px;" />
                </div>
            </form>
            %(footer)s
        </center>
    </body>
</html>
"""

# template for the file listing. variables:
# 'title'
page_template_begin = """\
<html>
    <head>
        <title>Simple Share - %(title)s</title>
    </head>
    <body style="font-family:sans-serif;color:black;background:#bbddff">
        <center>
            <div style="color:white;background:#556688;font-size:150%%">
                Simple Share
            </div>
            <h2>%(title)s</h2>
            <div>
"""
page_template_end = """\
            </div>
            <div>&nbsp;</div>
            <div style="color:white;background:#556688;font-size:150%%">
                &nbsp;
            </div>
            %(footer)s
        </center>
    </body>
</html>
"""

page_error_message = """\
<html>
    <head>
        <title>Simple Share - Error response</title>
    </head>
    <body
      style="font-family:sans-serif;color:black;background:#bbddff"
      onload="javascript:document.login.response.focus()"
    >
        <center>
            <div style="color:white;background:#556688;font-size:150%%">
                Simple Share
            </div>
            <h1>Error response</h1>
            <p>Error code <span style="font-size:150%%">%(code)d.</span></p>
            <p>Message: <span style="font-size:150%%">%(message)s</span>.</p>
            <p>Error code explanation: %(code)s = %(explain)s.</p>
            <p><a href="javascript:window.back()">Back</a></p>
            <div style="color:white;background:#556688;font-size:150%%">
                &nbsp;
            </div>
        </center>
    </body>
</html>
"""

# -----------------------------------------------------------------------------

class TokenSecuredHTTPRequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    """\
    This request handler manages clients with session cookies.
    If the client connects the first time it will be redirected to the login
    page.
    """
    # the sessions are tracked in this dictionary
    sessions = {}
    
    # override error message template
    error_message_format = page_error_message
    
    def __init__(self, *args):
        """instantiated on each request"""
        self.session_cookie = None
        SimpleHTTPServer.SimpleHTTPRequestHandler.__init__(self, *args)
    
    def is_local_access(self):
        """\
        Check if an access is comming from localhost.
        """
        # XXX this check only works for IPv4 sockets but that is fine for now
        # as the underlying SocketServer uses a v4 socket
        return self.client_address[0] == '127.0.0.1'
    
    def redirect(self, location):
        """helper to generate the headers for page redirection, caller must
           not output anything."""
        self.send_response(301)
        self.send_header("Location", location)
        self.end_headers()
        
    def page_session(self, active_session):
        """\
        Display details about one session.
        """
        if not self.is_local_access():
            active_session.track(self.path, 'denied')
            self.send_error(403)
            return
        
        request = cgi.parse_qs(self.query)
        # check id, if no id given, redirect to list
        if 'id' not in request:
            self.redirect('/sessions')
            return
        
        try:
            session = self.sessions[request['id'][0]]
        except KeyError:
            # invalid session, go to session list
            self.redirect('/sessions')
            return
        
        # check for commands
        if 'action' in request:
            for action in request['action']:
                if action == 'delete':
                    # remove session and go back to the session list
                    del self.sessions[session.id]
                    self.redirect('/sessions')
                    return
                elif action == 'accept':
                    session.track(self.path, 'accept')
                    session.access_granted = True
                elif action == 'deny':
                    session.track(self.path, 'deny')
                    session.access_granted = False
                elif action == 'list':
                    self.redirect('/sessions'); return
        
        # define a helper function that makes links to this page
        def link(message, action=None):
            if action is None:
                action = ''
            else:
                action = '&action=' + action
            return '<a href="/session?id=%s%s">%s</a>' % (session.id, action, message)
        
        # return page
        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        self.send_header('Cache-Control', 'no-store')
        self.end_headers()
        # html page intro
        self.wfile.write(page_template_begin % {'title': 'Session detail'})
        # mark if it's the session of the user
        if session == active_session:
            self.wfile.write("<div>This is your current session!</div>")
        # details
        self.wfile.write('<table border=0 cellspacing="5px" style="%s">\n' % css_border)
        for name, object, style in [
            ('Session ID',  session.id,         css_tt),
            ('Challange',   session.challange,  css_tt+css_challange),
            ('Response',    session.response,   css_tt+css_response),
            ('Last Address', ':'.join([str(x) for x in session.client_address]), ''),
            ('Access',  session.access_granted and link('yes', 'deny') or link('no', 'accept'), ''),
            ('Expires', format_time(session.expires), ''),
        ]:
            self.wfile.write('<tr><th align="left">%s</th> <td %s>%s</td> </tr>' % (
                name,
                style and 'style="%s"' % style or '',
                object
            ))
        self.wfile.write("</table>\n")
        # actions
        self.wfile.write('%s | %s | ' % (link('Refresh'), link('Delete', 'delete')))
        self.wfile.write('Back to the <a href="/sessions">session list</a></div>\n')
        # history
        self.wfile.write('<h3>Trail</h3>\n')
        self.wfile.write('<table border=0 cellspacing="5px" style="%s">\n' % css_border)
        for date, location, message in session.trail:
            self.wfile.write('<tr><td style="color:gray">%s</td> <td>%s</td> <td><a href="%s">%s</a></td> </tr>' % (
                format_time(date),
                message,
                location,
                location,
            ))
        self.wfile.write("</table>")
        # html page end
        self.wfile.write(page_template_end % {'footer': '<div style="font-size:75%">This page can only be viewed from localhost.</div>'})
    
    
    def page_sessions(self, active_session):
        """\
        Display a list of all active sessions with links to the detail page.
        """
        if not self.is_local_access():
            active_session.track(self.path, 'denied')
            self.send_error(403)
            return
        
        # return page
        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        self.send_header('Cache-Control', 'no-store')
        self.end_headers()
        
        self.wfile.write(page_template_begin % {'title': 'Session listing'})
        self.wfile.write('<div>The following is a list of sessions. Click on the ID for more details.</div>')
        self.wfile.write('<table border=0 cellspacing="5px" style="%s"><tr>\n' % css_border)
        for n, column in enumerate(['ID', 'Challange', 'Response', 'Address', 'Access', 'Expires']):
            self.wfile.write('<th align="left">%s</th> ' % column)
        self.wfile.write('</tr>\n')
        row_template = '<tr><td style="' + css_tt + '"><a href="session?id=%s">%s</a></td>' \
                       '<td style="' + css_tt + css_challange + '">%s</td>' \
                       '<td style="' + css_tt + css_response + '">%s</td>' \
                       '<td>%s</td> <td>%s</td> <td>%s</td> </tr>\n'
        for session in self.sessions.values():
            self.wfile.write(row_template % (
                session.id,
                session == active_session and "<strong>%s</strong>" % session.id or session.id,
                session.challange,
                session.response,
                ':'.join([str(x) for x in session.client_address]),
                session.access_granted
                    and '<a href="/session?id=%s&action=deny&action=list">yes</a>' % session.id
                    or '<a href="/session?id=%s&action=accept&action=list">no</a>' % session.id,
                format_time(session.expires),
            ))
        self.wfile.write("</table>")
        self.wfile.write('<div><a href="/sessions">Refresh</a> | Back to the <a href="/">root</a></div>')
        self.wfile.write(page_template_end % {'footer':'<div style="font-size:75%">This page can only be viewed from localhost.</div>'})
    
    
    def page_login(self, session):
        """\
        Display the the login page and check incoming form data.
        """
        # authenticated user wants login page, we're not allowing that
        # and send him to the root
        if session.access_granted: self.redirect('/'); return
        
        message = "Login Required."
        request = cgi.parse_qs(self.query)
        # ensure that redirection is set
        if 'redirect' not in request:
            request['redirect'] = ['/']
            # XXX or use referer
        # check for actual login
        if 'response' in request:
            #~ print "ZZZ checking", session.response, request['response'][0]
            if session.response == request['response'][0]:
                # login OK! go back the the page indicated by 'redirect'
                session.track(self.path, "login OK")
                session.access_granted = True
                self.redirect(request['redirect'][0])
                return
            else:
                message = "Login failed. Try again."
        # show login page
        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        self.send_header('Cache-Control', 'no-store')
        self.end_headers()
        self.wfile.write(page_login_form % {
            'redirect': request['redirect'][0],
            'challange': session.challange,
            'message': message,
            'footer': self.is_local_access() and '<a href="/sessions">Session managment</a>' or ' ',
        })


    def do_GET(self):
        """handle get requests, login is handled here, other requests are
           delegated to the super class impl of this method"""
        # housekeeping, remove old expired sessions
        for session_id, session in self.sessions.items():
            if session.is_expired():
                #~ print "PPP removing expired session", session
                del self.sessions[session_id]
        
        # check for login
        session = self.get_session()
        session.client_address = self.client_address #update last access address
        session.prolong()

        proto, host, self.file_part, self.query, fragment = urlparse.urlsplit(self.path)
        # check for "implemented" pages
        page_handler = 'page_%s' % self.file_part[1:]
        if hasattr(self, page_handler):
            session.track(self.path, "GET")
            # found a page handler function, call it
            getattr(self, page_handler)(session)
        elif session.access_granted:
            # other paths are files
            session.track(self.path, 'accessed')
            start = time.time()
            SimpleHTTPServer.SimpleHTTPRequestHandler.do_GET(self)
            end = time.time()
            session.track(self.path, 'done in %.2fs' % (end-start))
        else:
            # session not ok, require login
            session.track(self.path, "denied access")
            self.redirect('/login?redirect=%s' % urllib.quote(self.file_part))

    def do_HEAD(self):
        """same as GET, just w/o contents"""
        session = self.get_session()
        session.track(self.path, "HEAD")
        if self.check_session(session):
            SimpleHTTPServer.SimpleHTTPRequestHandler.do_HEAD(self)
        else:
            self.send_error(401)

    def end_headers(self):
        """override this to append the session cookie"""
        if self.session_cookie is not None:
            self.wfile.write(self.session_cookie)
            self.wfile.write('\r\n')
        SimpleHTTPServer.SimpleHTTPRequestHandler.end_headers(self)

    def get_session(self):
        """get the cookies from the request headers and find an open session
           or greate a new session if none is found. the session object is
           returned."""
        #~ print self.raw_requestline
        #~ print self.path
        #~ print self.headers
        # parse cookies
        self.cookies = Cookie.SimpleCookie(', '.join(filter(None, self.headers.getheaders('cookie'))))
        
        # try to load the session based on the cookie
        if self.cookies.has_key('simple_share') and self.cookies['simple_share'].value in self.sessions:
            session = self.sessions[self.cookies['simple_share'].value]
        else:
            # if we don't have a session, make a new one
            # in that case, we're setting the cookie
            session = Session()
            self.sessions[session.id] = session
            self.session_cookie = Cookie.SimpleCookie()
            self.session_cookie['simple_share'] = session.id
            #~ self.session_cookie['simple_share']['expires'] = int(session.expires - time.time())
        return session

    # must override this to get the own style
    def list_directory(self, path):
        """Helper to produce a directory listing (absent index.html).

        Return value is either a file object, or None (indicating an
        error).  In either case, the headers are sent, making the
        interface the same as for send_head().
        """
        try:
            list = os.listdir(path)
        except os.error:
            self.send_error(404, "No permission to list directory")
            return None
        list.sort(key=lambda a: a.lower())
        f = StringIO()
        displaypath = cgi.escape(urllib.unquote(self.path))
        f.write(page_template_begin % {'title': 'Directory listing for %s' % displaypath})
        f.write('<table border=0 cellspacing="5px" style="%s">\n' % css_border)
        f.write('<tr><th align="left">Name</th> <th align="left">Type/Size</th></tr>\n')
        if displaypath != '/':
            # non root, show '..' parent
            parent, _ = posixpath.split(displaypath[:-1]) #cut off trailing '/'
            f.write('<tr><td><a href="%s">%s</a></td> <td>%s</td></tr>\n' % (
                parent,
                '..',
                'parent'
            ))
        file_count = 0
        dir_count = 0
        for name in list:
            fullname = os.path.join(path, name)
            statinfo = os.stat(fullname)
            displayname = linkname = name
            # Append / for directories or @ for symbolic links
            if stat.S_ISDIR(statinfo.st_mode):
                displayname = name + "/"
                linkname = name + "/"
                dir_count += 1
            else:
                file_count += 1
            if stat.S_ISLNK(statinfo.st_mode):
                displayname = name + "@"
                # Note: a link to a directory displays with @ and links with /
            f.write('<tr><td><a href="%s">%s</a></td> <td>%s</td></tr>\n' % (
                urllib.quote(linkname),
                cgi.escape(displayname),
                stat.S_ISDIR(statinfo.st_mode) and 'directory' or bytes2string(statinfo.st_size),
            ))
        f.write("</table>\n<div>")
        # stats
        if file_count:
            f.write("%d file%s " % (file_count, file_count > 1 and 's' or ''))
        if dir_count:
            f.write("%d director%s" % (dir_count, dir_count > 1 and 'ies' or 'y'))
        if not file_count and not dir_count:
            f.write("No items found.")
        f.write("</div>\n")
        
        f.write(page_template_end % {
            'footer': self.is_local_access() and '\n<div><a href="/sessions">Session managment</a></div>\n' or ''
        })
        length = f.tell()
        f.seek(0)
        self.send_response(200)
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(length))
        self.end_headers()
        return f

# -----------------------------------------------------------------------------

class ThreadingHTTPServer(BaseHTTPServer.HTTPServer, SocketServer.ThreadingMixIn):
    pass

# -----------------------------------------------------------------------------
# simple main
# -----------------------------------------------------------------------------
if __name__ == '__main__':
    # try to find the IP numbers of our network interfaces
    if sys.platform == 'linux2':
        print 'IPs:', ' '.join([
            l.split()[1]
            for l in os.popen('LANG=C /sbin/ifconfig').read().splitlines()
            if l.strip().startswith('inet ')
        ])
    elif sys.platform == 'win32':
        print 'IPs:', ' '.join([
            l.split(':')[1].strip()
            for l in os.popen("ipconfig").read().splitlines()
            if '  IP' in l
        ])
    # run the web server.
    try:
        SimpleHTTPServer.test(
            HandlerClass = TokenSecuredHTTPRequestHandler,
            ServerClass = ThreadingHTTPServer
        )
    except KeyboardInterrupt:
        # Keyboard interrupt (CTRL+C) is used to exit the program and
        # should not generate an error
        pass
